"""rename vegetation columns

Revision ID: 0219
Revises: 0218
Create Date: 2024-01-22 15:00

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '0219'
down_revision = '0218'
branch_labels = None
depends_on = None

MIGRATION_QUERIES = """
SELECT RecoverGeometryColumn('v2_cross_section_location', 'the_geom', 4326, 'POINT', 'XY')
"""


def upgrade():
    with op.batch_alter_table("v2_cross_section_location") as batch_op:
        batch_op.alter_column("friction_value", nullable=True, type_=sa.FLOAT)
    for q in MIGRATION_QUERIES.split(";"):
        op.execute(sa.text(q))


def downgrade():
    with op.batch_alter_table("v2_cross_section_location") as batch_op:
        batch_op.alter_column("friction_value", nullable=False, type_=sa.Float)



import struct

import pytest

from threedi_schema import ModelSchema, ThreediDatabase

# from threedi_schema.migrations.versions import 0213_connected_points_to_breaches
module_name = "0213_connected_points_to_breaches"
migration_213 = getattr(
    __import__("threedi_schema.migrations.versions", fromlist=[module_name]),
    module_name,
)

CalculationPoint = migration_213.CalculationPoint
ConnectedPoint = migration_213.ConnectedPoint
Levee = migration_213.Levee
PotentialBreach = migration_213.PotentialBreach
ConnectionNode = migration_213.ConnectionNode
Manhole = migration_213.Manhole
Channel = migration_213.Channel


@pytest.fixture
def sqlite_v212():
    """An in-memory database with schema version 212"""
    db = ThreediDatabase("")
    ModelSchema(db).upgrade("0212", backup=False)
    return db


@pytest.fixture
def session(sqlite_v212):
    return sqlite_v212.get_session()


def parse_hexewkb(ewkb):
    """Extract 4 line coordinates from an ewkb"""
    return struct.unpack("4d", bytes.fromhex(ewkb)[-32:])


GEOM1 = "SRID=4326;POINT (0 0)"
GEOM2 = "SRID=4326;POINT (0 1)"
CHANNEL = "SRID=4326;LINESTRING(0 0,10 0)"
CHANNEL_INV = "SRID=4326;LINESTRING(10 0,0 0)"
LINE = "0102000020E610000002000000000000000000000000000000000000000000000000000000000000000000F03F"


@pytest.mark.parametrize(
    "objs",
    [
        [CalculationPoint(id=1, the_geom=GEOM1)],
        [ConnectedPoint(id=1, the_geom=GEOM1, calculation_pnt_id=None)],
        [ConnectedPoint(id=1, the_geom=GEOM1, calculation_pnt_id=1)],
        [
            CalculationPoint(id=1, the_geom=GEOM1),
            ConnectedPoint(id=1, the_geom=GEOM1, calculation_pnt_id=1),
        ],
        [
            CalculationPoint(id=1, the_geom=GEOM1),
            ConnectedPoint(id=1, the_geom=GEOM1, calculation_pnt_id=1),
            ConnectedPoint(id=2, the_geom=GEOM1, calculation_pnt_id=1),
        ],
    ],
)
def test_clean_connected_points(session, objs):
    session.add_all(objs)
    session.flush()
    migration_213.clean_connected_points(session)
    assert session.query(CalculationPoint).count() == 0
    assert session.query(ConnectedPoint).count() == 0


@pytest.mark.parametrize(
    "objs",
    [
        [
            CalculationPoint(id=1, the_geom=GEOM1),
            ConnectedPoint(id=1, the_geom=GEOM2, calculation_pnt_id=1),
        ],
        [
            CalculationPoint(id=1, the_geom=GEOM1),
            ConnectedPoint(id=1, the_geom=GEOM2, calculation_pnt_id=1),
            ConnectedPoint(id=2, the_geom=GEOM2, calculation_pnt_id=1),
        ],
        [
            CalculationPoint(id=1, the_geom=GEOM1),
            ConnectedPoint(
                id=1, the_geom=GEOM2, calculation_pnt_id=1, exchange_level=1.0
            ),
        ],
        [
            CalculationPoint(id=1, the_geom=GEOM1),
            ConnectedPoint(id=1, the_geom=GEOM2, calculation_pnt_id=1, levee_id=1),
        ],
    ],
)
def test_clean_connected_points_keep(session, objs):
    session.add_all(objs)
    session.flush()
    migration_213.clean_connected_points(session)

    actual = (
        session.query(CalculationPoint).count() + session.query(ConnectedPoint).count()
    )
    assert actual == len(objs)


def todict(x):
    return {col.name: getattr(x, col.name) for col in x.__table__.columns}


def assert_sqlalchemy_objects_equal(a, b):
    assert a.__class__ is b.__class__
    if a is None:
        return
    assert todict(a) == todict(b)


@pytest.mark.parametrize(
    "objs,expected",
    [
        [
            [
                CalculationPoint(id=1, user_ref="123#4#v2_channel#1"),
                ConnectedPoint(id=1, the_geom=GEOM2, calculation_pnt_id=1),
                Channel(id=4, the_geom=CHANNEL),
            ],
            PotentialBreach(channel_id=4, the_geom=LINE, code="1#123#4#v2_channel#1"),
        ],
        [
            [
                CalculationPoint(id=1, user_ref="123#4#v2_channel#1"),
                ConnectedPoint(
                    id=1, the_geom=GEOM2, calculation_pnt_id=1, exchange_level=1.1
                ),
                Channel(id=4, the_geom=CHANNEL),
            ],
            PotentialBreach(
                channel_id=4,
                the_geom=LINE,
                exchange_level=1.1,
                code="1#123#4#v2_channel#1",
            ),
        ],
        [
            [
                CalculationPoint(id=1, user_ref="123#4#v2_channel#1"),
                ConnectedPoint(id=1, the_geom=GEOM2, calculation_pnt_id=1, levee_id=4),
                Levee(id=4, crest_level=1.1),
                Channel(id=4, the_geom=CHANNEL),
            ],
            PotentialBreach(
                channel_id=4,
                the_geom=LINE,
                exchange_level=1.1,
                code="1#123#4#v2_channel#1",
            ),
        ],
        [
            [
                CalculationPoint(id=1, user_ref="123#4#v2_channel#1"),
                ConnectedPoint(
                    id=1,
                    the_geom=GEOM2,
                    calculation_pnt_id=1,
                    exchange_level=1.1,
                    levee_id=4,
                ),
                Levee(id=4, crest_level=1.2, max_breach_depth=0.5, material=1),
                Channel(id=4, the_geom=CHANNEL),
            ],
            PotentialBreach(
                channel_id=4,
                the_geom=LINE,
                exchange_level=1.1,
                maximum_breach_depth=0.5,
                levee_material=1,
                code="1#123#4#v2_channel#1",
            ),
        ],
        [
            [
                CalculationPoint(id=1, user_ref="123#3#v2_manhole#1"),
                ConnectedPoint(id=1, the_geom=GEOM2, calculation_pnt_id=1),
                Manhole(id=3, connection_node_id=6),
                ConnectionNode(id=6),
                Channel(
                    id=4,
                    the_geom=CHANNEL,
                    connection_node_start_id=6,
                    calculation_type=102,
                ),
            ],
            PotentialBreach(channel_id=4, the_geom=LINE, code="1#123#3#v2_manhole#1"),
        ],
        [
            [
                CalculationPoint(id=1, user_ref="123#3#v2_manhole#1"),
                ConnectedPoint(id=1, the_geom=GEOM2, calculation_pnt_id=1),
                Manhole(id=3, connection_node_id=6),
                ConnectionNode(id=6),
                Channel(
                    id=4,
                    the_geom=CHANNEL_INV,
                    connection_node_end_id=6,
                    calculation_type=102,
                ),
            ],
            PotentialBreach(channel_id=4, the_geom=LINE, code="1#123#3#v2_manhole#1"),
        ],
        [
            [
                CalculationPoint(id=1, user_ref="123#3#v2_manhole#1"),
                ConnectedPoint(id=1, the_geom=GEOM2, calculation_pnt_id=1),
                Manhole(id=3, connection_node_id=6),
                ConnectionNode(id=6),
                Channel(
                    id=4,
                    the_geom=CHANNEL,
                    connection_node_start_id=6,
                    calculation_type=105,
                ),
            ],
            PotentialBreach(channel_id=4, the_geom=LINE, code="1#123#3#v2_manhole#1"),
        ],
        [
            [
                CalculationPoint(id=1, user_ref="123#3#v2_manhole#1"),
                ConnectedPoint(id=1, the_geom=GEOM2, calculation_pnt_id=1),
                Manhole(id=3, connection_node_id=6),
                ConnectionNode(id=6),
            ],
            None,
        ],
        [
            [
                CalculationPoint(id=1, user_ref="123#3#v2_manhole#1"),
                ConnectedPoint(id=1, the_geom=GEOM2, calculation_pnt_id=1),
                Manhole(id=3, connection_node_id=6),
                ConnectionNode(id=6),
                Channel(id=4, connection_node_start_id=6, calculation_type=101),
            ],
            None,
        ],
        [
            [
                CalculationPoint(id=1, user_ref="123#3#v2_manhole#1"),
                ConnectedPoint(id=1, the_geom=GEOM2, calculation_pnt_id=1),
                Manhole(id=3, connection_node_id=6),
                ConnectionNode(id=6),
                Channel(id=3, connection_node_start_id=6, calculation_type=102),
                Channel(id=5, connection_node_start_id=6, calculation_type=105),
                Channel(
                    id=4,
                    the_geom=CHANNEL_INV,
                    connection_node_end_id=6,
                    calculation_type=105,
                ),
            ],
            PotentialBreach(channel_id=4, the_geom=LINE, code="1#123#3#v2_manhole#1"),
        ],
    ],
)
def test_to_potential_breach(session, objs, expected):
    session.add_all(objs)
    session.flush()
    actual = migration_213.to_potential_breach(session, 1)

    assert_sqlalchemy_objects_equal(actual, expected)


@pytest.mark.parametrize(
    "node_idx,calc_pnt_x,calc_pnt_y,x,y",
    [
        (0, 0, 0, 0, 0),
        (-1, 0, 0, 10, 10),
        (1, 0, 5, 0, 5),
        (1, 0, 10, 0, 10),
        (1, 0, 10 + 1e-8, 0, 10),
        (2, 10 + 7e-8, 10 - 7e-8, 10, 10),
    ],
)
def test_get_breach_line_geom(session, node_idx, calc_pnt_x, calc_pnt_y, x, y):
    objs = [
        ConnectedPoint(id=1, the_geom="SRID=4326;POINT (10 0)", calculation_pnt_id=2),
        CalculationPoint(id=2, the_geom=f"SRID=4326;POINT({calc_pnt_x} {calc_pnt_y})"),
        Channel(id=3, the_geom="SRID=4326;LINESTRING(0 0,0 10,10 10)"),
    ]

    session.add_all(objs)
    session.flush()
    geom = migration_213.get_breach_line_geom(session, 1, 3, node_idx)
    x1, y1, x2, y2 = parse_hexewkb(geom)
    assert x1 == x
    assert y1 == y
    assert x2 == 10.0
    assert y2 == 0.0



import sqlite3

import pytest

from threedi_schema import ModelSchema
from threedi_schema.migrations.exceptions import InvalidSRIDException


@pytest.mark.parametrize("epsg_code", [
    999999, # non-existing
    2227, # projected / US survey foot
    4979, # not project
])
def test_check_valid_crs(in_memory_sqlite, epsg_code):
    schema = in_memory_sqlite.schema
    schema.upgrade(revision="0229", backup=False)
    schema._set_custom_epsg_code(epsg_code)
    with pytest.raises(InvalidSRIDException) as exc_info:
        schema.upgrade(backup=False)


def test_migration(tmp_path_factory, oldest_sqlite):
    schema = ModelSchema(oldest_sqlite)
    schema.upgrade(backup=False)
    cursor = sqlite3.connect(schema.db.path).cursor()
    query = cursor.execute("SELECT srid FROM geometry_columns where f_table_name = 'geom'")
    epsg_matches = [int(item[0])==28992 for item in query.fetchall()]
    assert all(epsg_matches)





"""remove vegetation_drag_coeficients column

Revision ID: 0221
Revises: 0220
Create Date: 2024-02-04

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '0221'
down_revision = '0220'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("v2_cross_section_location") as batch_op:
        batch_op.drop_column("vegetation_drag_coeficients")
    op.execute(sa.text("SELECT RecoverGeometryColumn('v2_cross_section_location', 'the_geom', 4326, 'POINT', 'XY')"))


def downgrade():
    # No downgrade implemented because vegetation_drag_coeficients should
    # never have been addd to v2_cross_section_location
    pass



class InvalidSRIDException(Exception):
    def __init__(self, epsg_code, issue=None):
        msg = f"Cannot migrate schematisation with model_settings.epsg_code={epsg_code}"
        if issue is not None:
            msg += f"; {issue}"
        super().__init__(msg)


from typing import List

import sqlalchemy as sa


def drop_geo_table(op, table_name: str):
    """

    Safely drop table, taking into account geometry columns

    Parameters:
    op : object
        An object representing the database operation.
    table_name : str
        The name of the table to be dropped.
    """
    op.execute(sa.text(f"SELECT DropTable(NULL, '{table_name}');"))


def drop_conflicting(op, new_tables: List[str]):
    """
    Drop tables from database that conflict with new tables

    Parameters:
    op: The SQLAlchemy operation context to interact with the database.
    new_tables: A list of new table names to be checked for conflicts with existing tables.
    """
    connection = op.get_bind()
    existing_tables = [item[0] for item in connection.execute(
        sa.text("SELECT name FROM sqlite_master WHERE type='table';")).fetchall()]
    for table_name in set(existing_tables).intersection(new_tables):
        drop_geo_table(op, table_name)


"""Upgrade settings in schema

Revision ID: 0225
Revises:
Create Date: 2024-08-30 07:52

"""
from typing import Dict, List, Tuple

import sqlalchemy as sa
from alembic import op
from sqlalchemy import Boolean, Column, Float, Integer, String, Text
from sqlalchemy.orm import declarative_base

from threedi_schema.migrations.utils import drop_conflicting, drop_geo_table

# revision identifiers, used by Alembic.
revision = "0226"
down_revision = "0225"
branch_labels = None
depends_on = None

RENAME_TABLES = [
    ("v2_dem_average_area", "dem_average_area"),
    ("v2_exchange_line", "exchange_line"),
    ("v2_grid_refinement", "grid_refinement_line"),
    ("v2_grid_refinement_area", "grid_refinement_area"),
    ("v2_obstacle", "obstacle"),
    ("v2_potential_breach", "potential_breach"),
]

NEW_COLUMNS = [
    ("dem_average_area", Column("tags", Text)),
    ("dem_average_area", Column("code", Text)),
    ("dem_average_area", Column("display_name", Text)),
    ("exchange_line", Column("tags", Text)),
    ("exchange_line", Column("code", Text)),
    ("exchange_line", Column("display_name", Text)),
    ("grid_refinement_line", Column("tags", Text)),
    ("grid_refinement_area", Column("tags", Text)),
    ("obstacle", Column("tags", Text)),
    ("obstacle", Column("display_name", Text)),
    ("potential_breach", Column("tags", Text)),
    ("potential_breach", Column("final_exchange_level", Float))
]

RENAME_COLUMNS = {
    "grid_refinement_line": {"refinement_level": "grid_level"},
    "grid_refinement_area": {"refinement_level": "grid_level"},
    "potential_breach": {"exchange_level": "initial_exchange_level"}
}

RETYPE_COLUMNS = {
    "potential_breach": [("channel_id", "INTEGER")],
    "exchange_line": [("channel_id", "INTEGER")],
}

REMOVE_COLUMNS = {
    "exchange_line": ["channel"],
    "potential_breach": ["channel", "maximum_breach_depth"]
}


class Schema225UpgradeException(Exception):
    pass


def add_columns_to_tables(table_columns: List[Tuple[str, Column]]):
    # no checks for existence are done, this will fail if any column already exists
    for dst_table, col in table_columns:
        with op.batch_alter_table(dst_table) as batch_op:
            batch_op.add_column(col)


def remove_tables(tables: List[str]):
    for table in tables:
        drop_geo_table(op, table)


def modify_table(old_table_name, new_table_name):
    # Create a new table named `new_table_name` by copying the
    # data from `old_table_name`.
    # Use the columns from `old_table_name`, with the following exceptions:
    # * columns in `REMOVE_COLUMNS[new_table_name]` are skipped
    # * columns in `RENAME_COLUMNS[new_table_name]` are renamed
    # * columns in `RETYPE_COLUMNS[new_table_name]` change type
    # * `the_geom` is renamed to `geom` and NOT NULL is enforced
    connection = op.get_bind()
    columns = connection.execute(sa.text(f"PRAGMA table_info('{old_table_name}')")).fetchall()
    # get all column names and types
    col_names = [col[1] for col in columns]
    col_types = [col[2] for col in columns]
    # get type of the geometry column
    geom_type = None
    for col in columns:
        if col[1] == 'the_geom':
            geom_type = col[2]
            break
    # create list of new columns and types for creating the new table
    # create list of old columns to copy to new table
    skip_cols = ['id', 'the_geom']
    if new_table_name in REMOVE_COLUMNS:
        skip_cols += REMOVE_COLUMNS[new_table_name]
    old_col_names = []
    new_col_names = []
    new_col_types = []
    for cname, ctype in zip(col_names, col_types):
        if cname in skip_cols:
            continue
        old_col_names.append(cname)
        if new_table_name in RENAME_COLUMNS and cname in RENAME_COLUMNS[new_table_name]:
            new_col_names.append(RENAME_COLUMNS[new_table_name][cname])
        else:
            new_col_names.append(cname)
        if new_table_name in RETYPE_COLUMNS and cname in RETYPE_COLUMNS[new_table_name]:
            new_col_types.append(RETYPE_COLUMNS[new_table_name][cname])
        else:
            new_col_types.append(ctype)
    # add to the end manually
    old_col_names.append('the_geom')
    new_col_names.append('geom')
    new_col_types.append(f'{geom_type} NOT NULL')
    # Create new table (temp), insert data, drop original and rename temp to table_name
    new_col_str = ','.join(['id INTEGER PRIMARY KEY NOT NULL'] + [f'{cname} {ctype}' for cname, ctype in
                                                                  zip(new_col_names, new_col_types)])
    op.execute(sa.text(f"CREATE TABLE {new_table_name} ({new_col_str});"))
    # Copy data
    op.execute(sa.text(f"INSERT INTO {new_table_name} (id, {','.join(new_col_names)}) "
                       f"SELECT id, {','.join(old_col_names)} FROM {old_table_name}"))


def fix_geometry_columns():
    GEO_COL_INFO = [
        ('dem_average_area', 'geom', 'POLYGON'),
        ('exchange_line', 'geom', 'LINESTRING'),
        ('grid_refinement_line', 'geom', 'LINESTRING'),
        ('grid_refinement_area', 'geom', 'POLYGON'),
        ('obstacle', 'geom', 'LINESTRING'),
        ('potential_breach', 'geom', 'LINESTRING'),
    ]
    for table, column, geotype in GEO_COL_INFO:
        migration_query = f"SELECT RecoverGeometryColumn('{table}', '{column}', {4326}, '{geotype}', 'XY')"
        op.execute(sa.text(migration_query))


def set_potential_breach_final_exchange_level():
    conn = op.get_bind()
    res = conn.execute(sa.text(
        """
        SELECT id FROM v2_potential_breach
        WHERE exchange_level IS NOT NULL AND maximum_breach_depth IS NULL;
        """
    )).fetchall()
    if len(res) > 0:
        raise Schema225UpgradeException(
            f"Could not set final_exchange_level because maximum_breach_depth was not defined for rows: {res}")
    op.execute(sa.text(
        """
        UPDATE potential_breach
        SET final_exchange_level = (
            SELECT vb.exchange_level - vb.maximum_breach_depth
            FROM v2_potential_breach vb
            WHERE vb.id = potential_breach.id
            AND exchange_level IS NOT NULL
        );
        """
    ))


def upgrade():
    # Drop tables that conflict with new table names
    drop_conflicting(op, [new_name for _, new_name in RENAME_TABLES])
    rem_tables = []
    for old_table_name, new_table_name in RENAME_TABLES:
        modify_table(old_table_name, new_table_name)
        rem_tables.append(old_table_name)
    add_columns_to_tables(NEW_COLUMNS)
    set_potential_breach_final_exchange_level()
    fix_geometry_columns()
    remove_tables(rem_tables)


def downgrade():
    # Not implemented on purpose
    raise NotImplementedError("Downgrade back from 0.3xx is not supported")



"""Upgrade settings in schema

Revision ID: 0225
Revises:
Create Date: 2024-09-10 09:00

"""
import csv
import uuid
from pathlib import Path
from typing import Dict, List, Tuple

import sqlalchemy as sa
from alembic import op
from sqlalchemy import Column, Float, func, Integer, select, String, Text
from sqlalchemy.orm import declarative_base, Session

from threedi_schema.domain import constants
from threedi_schema.domain.custom_types import Geometry, IntegerEnum
from threedi_schema.migrations.utils import drop_conflicting, drop_geo_table

Base = declarative_base()

data_dir = Path(__file__).parent / "data"


# revision identifiers, used by Alembic.
revision = "0228"
down_revision = "0227"
branch_labels = None
depends_on = None

RENAME_TABLES = [
    ("v2_channel", "channel"),
    ("v2_windshielding", "windshielding_1d"),
    ("v2_cross_section_location", "cross_section_location"),
    ("v2_pipe", "pipe"),
    ("v2_culvert", "culvert"),
    ("v2_weir", "weir"),
    ("v2_orifice", "orifice"),
    ("v2_pumpstation", "pump")
]

DELETE_TABLES = ["v2_cross_section_definition",
                 "v2_floodfill",
                 "v2_connection_nodes"]

RENAME_COLUMNS = {
    "culvert": {"calculation_type": "exchange_type",
                "dist_calc_points": "calculation_point_distance",
                "invert_level_start_point": "invert_level_start",
                "invert_level_end_point": "invert_level_end",
                "connection_node_start_id": "connection_node_id_start",
                "connection_node_end_id": "connection_node_id_end"},
    "pipe": {"calculation_type": "exchange_type",
             "dist_calc_points": "calculation_point_distance",
             "material": "material_id",
                "invert_level_start_point": "invert_level_start",
                "invert_level_end_point": "invert_level_end",
                "connection_node_start_id": "connection_node_id_start",
                "connection_node_end_id": "connection_node_id_end"},
    "channel": {"calculation_type": "exchange_type",
                "dist_calc_points": "calculation_point_distance",
                "connection_node_start_id": "connection_node_id_start",
                "connection_node_end_id": "connection_node_id_end"},
    "weir": {"calculation_type": "exchange_type",
                "dist_calc_points": "calculation_point_distance",
                "connection_node_start_id": "connection_node_id_start",
                "connection_node_end_id": "connection_node_id_end"},
    "orifice": {"calculation_type": "exchange_type",
                "dist_calc_points": "calculation_point_distance",
                "connection_node_start_id": "connection_node_id_start",
                "connection_node_end_id": "connection_node_id_end"},
    "pump": {"connection_node_start_id": "connection_node_id"}
}

REMOVE_COLUMNS = {
    "channel": ["zoom_category",],
    "cross_section_location": ["definition_id", "vegetation_drag_coeficients"],
    "culvert": ["zoom_category", "cross_section_definition_id"],
    "pipe": ["zoom_category", "original_length", "cross_section_definition_id", "profile_num"],
    "orifice": ["zoom_category", "cross_section_definition_id"],
    "weir": ["zoom_category", "cross_section_definition_id"],
    "pump": ["connection_node_end_id", "zoom_category", "classification"]
}

ADD_COLUMNS = [
    ("channel", Column("tags", Text)),
    ("cross_section_location", Column("tags", Text)),
    ("culvert", Column("tags", Text)),
    ("culvert", Column("material_id", Integer)),
    ("orifice", Column("tags", Text)),
    ("orifice", Column("material_id", Integer)),
    ("pipe", Column("tags", Text)),
    ("pump", Column("tags", Text)),
    ("weir", Column("tags", Text)),
    ("weir", Column("material_id", Integer)),
    ("windshielding_1d", Column("tags", Text)),
]

RETYPE_COLUMNS = {}


def add_columns_to_tables(table_columns: List[Tuple[str, Column]]):
    # no checks for existence are done, this will fail if any column already exists
    for dst_table, col in table_columns:
        if isinstance(col.type, Geometry):
            add_geometry_column(dst_table, col)
        else:
            with op.batch_alter_table(dst_table) as batch_op:
                batch_op.add_column(col)


def add_geometry_column(table: str, geocol: Column):
    # Adding geometry columns via alembic doesn't work
    # https://postgis.net/docs/AddGeometryColumn.html
    geotype = geocol.type
    query = (
        f"SELECT AddGeometryColumn('{table}', '{geocol.name}', {geotype.srid}, '{geotype.geometry_type}', 'XY', 1);")
    op.execute(sa.text(query))


class Schema228UpgradeException(Exception):
    pass


def add_columns_to_tables(table_columns: List[Tuple[str, Column]]):
    # no checks for existence are done, this will fail if any column already exists
    for dst_table, col in table_columns:
        with op.batch_alter_table(dst_table) as batch_op:
            batch_op.add_column(col)


def remove_tables(tables: List[str]):
    for table in tables:
        drop_geo_table(op, table)


def get_geom_type(table_name, geo_col_name):
    connection = op.get_bind()
    columns = connection.execute(sa.text(f"PRAGMA table_info('{table_name}')")).fetchall()
    for col in columns:
        if col[1] == geo_col_name:
            return col[2]

def modify_table(old_table_name, new_table_name):
    # Create a new table named `new_table_name` by copying the
    # data from `old_table_name`.
    # Use the columns from `old_table_name`, with the following exceptions:
    # * columns in `REMOVE_COLUMNS[new_table_name]` are skipped
    # * columns in `RENAME_COLUMNS[new_table_name]` are renamed
    # * columns in `RETYPE_COLUMNS[new_table_name]` change type
    # * `the_geom` is renamed to `geom` and NOT NULL is enforced
    connection = op.get_bind()
    columns = connection.execute(sa.text(f"PRAGMA table_info('{old_table_name}')")).fetchall()
    # get all column names and types
    col_names = [col[1] for col in columns]
    col_types = [col[2] for col in columns]
    # get type of the geometry column
    geom_type = get_geom_type(old_table_name, 'the_geom')
    # create list of new columns and types for creating the new table
    # create list of old columns to copy to new table
    skip_cols = ['id', 'the_geom']
    if new_table_name in REMOVE_COLUMNS:
        skip_cols += REMOVE_COLUMNS[new_table_name]
    old_col_names = []
    new_col_names = []
    new_col_types = []
    for cname, ctype in zip(col_names, col_types):
        if cname in skip_cols:
            continue
        old_col_names.append(cname)
        if new_table_name in RENAME_COLUMNS and cname in RENAME_COLUMNS[new_table_name]:
            new_col_names.append(RENAME_COLUMNS[new_table_name][cname])
        else:
            new_col_names.append(cname)
        if new_table_name in RETYPE_COLUMNS and cname in RETYPE_COLUMNS[new_table_name]:
            new_col_types.append(RETYPE_COLUMNS[new_table_name][cname])
        else:
            new_col_types.append(ctype)
    # add to the end manually
    old_col_names.append('the_geom')
    new_col_names.append('geom')
    new_col_types.append(f'{geom_type} NOT NULL')
    # Create new table (temp), insert data, drop original and rename temp to table_name
    new_col_str = ','.join(['id INTEGER PRIMARY KEY NOT NULL'] + [f'{cname} {ctype}' for cname, ctype in
                                                                  zip(new_col_names, new_col_types)])
    op.execute(sa.text(f"CREATE TABLE {new_table_name} ({new_col_str});"))
    # Copy data
    op.execute(sa.text(f"INSERT INTO {new_table_name} (id, {','.join(new_col_names)}) "
                       f"SELECT id, {','.join(old_col_names)} FROM {old_table_name}"))


def fix_geometry_columns():
    tables = ['channel', 'connection_node', 'cross_section_location', 'culvert',
              'orifice', 'pipe', 'pump', 'pump_map', 'weir', 'windshielding_1d']
    for table in tables:
        geom_type = get_geom_type(table, geo_col_name='geom')
        op.execute(sa.text(f"SELECT RecoverGeometryColumn('{table}', "
                           f"'geom', {4326}, '{geom_type}', 'XY')"))
        op.execute(sa.text(f"SELECT CreateSpatialIndex('{table}', 'geom')"))


class Temp(Base):
    __tablename__ = f'_temp_228_{uuid.uuid4().hex}'

    id = Column(Integer, primary_key=True)
    cross_section_table = Column(String)
    cross_section_friction_values = Column(String)
    cross_section_vegetation_table = Column(String)
    cross_section_shape = Column(IntegerEnum(constants.CrossSectionShape))
    cross_section_width = Column(Float)
    cross_section_height = Column(Float)


def extend_cross_section_definition_table():
    conn = op.get_bind()
    session = Session(bind=op.get_bind())
    # create temporary table
    op.execute(sa.text(
        f"""CREATE TABLE {Temp.__tablename__} 
            (id INTEGER PRIMARY KEY, 
             cross_section_table TEXT,
             cross_section_shape INT,
             cross_section_width REAL,
             cross_section_height REAL,
             cross_section_friction_values TEXT,
             cross_section_vegetation_table TEXT)
        """))
    res = conn.execute(sa.text("SELECT id, shape, width, height FROM v2_cross_section_definition")).fetchall()
    # process data from v2_cross_section_definition by setting width and height to None when it's not a single float
    # omitting this stop results in issues with the data types in the database
    data_to_insert = []
    for row in res:
        id, shape, width, height = row
        try:
            width = float(width)
        except (ValueError, TypeError):
            width = None
        try:
            height = float(height)
        except (ValueError, TypeError):
            height = None
        data_to_insert.append({"id": id, "shape": shape, "width": width, "height": height})    # insert data into the temp table
    for data in data_to_insert:
        conn.execute(sa.text(
            f"""INSERT INTO {Temp.__tablename__} (id, cross_section_shape, cross_section_width, cross_section_height)
            VALUES (:id, :shape, :width, :height)"""), data)  # Pass parameters as dictionary directly
    def make_table(*args):
        split_args = [arg.split() for arg in args]
        if not all(len(args) == len(split_args[0]) for args in split_args):
            return
        return '\n'.join([','.join(row) for row in zip(*split_args)])
    # Create cross_section_table for tabulated
    res = conn.execute(sa.text(f"""
        SELECT id, height, width, shape FROM v2_cross_section_definition 
        WHERE v2_cross_section_definition.shape IN (5,6,7)   
        AND height IS NOT NULL AND width IS NOT NULL
    """)).fetchall()
    update_data = []
    for id, h, w, s in res:
        # tabulated_YZ: width -> Y; height -> Z
        if s == constants.CrossSectionShape.TABULATED_YZ.value:
            cross_section_table = make_table(w, h)
        # tabulated_trapezium or tabulated_rectangle: height, width
        else:
            cross_section_table = make_table(h, w)
        update_data.append({"id": id, "cross_section_table": cross_section_table})
    session.bulk_update_mappings(Temp, update_data)
    session.commit()
    # add cross_section_friction_table to cross_section_definition
    res = conn.execute(sa.text("""
        SELECT id, friction_values FROM v2_cross_section_definition 
        WHERE friction_values IS NOT NULL
        AND v2_cross_section_definition.shape = 7 
    """)).fetchall()
    for id, friction_values in res:
        temp_row = session.query(Temp).filter_by(id=id).first()
        temp_row.cross_section_friction_values = friction_values.replace(' ',',')
        session.commit()
    # add cross_section_vegetation_table to cross_section_definition
    res = conn.execute(sa.text("""
        SELECT id, vegetation_stem_densities, vegetation_stem_diameters, vegetation_heights, vegetation_drag_coefficients
        FROM v2_cross_section_definition 
        WHERE vegetation_stem_densities IS NOT NULL
        AND vegetation_stem_diameters IS NOT NULL
        AND vegetation_heights IS NOT NULL
        AND v2_cross_section_definition.shape = 7 
        AND vegetation_drag_coefficients IS NOT NULL
    """)).fetchall()
    for id, dens, diam, h, c in res:
        temp_row = session.query(Temp).filter_by(id=id).first()
        temp_row.cross_section_vegetation_table = make_table(dens, diam, h, c)
        session.commit()


def migrate_cross_section_definition_from_temp(target_table: str,
                                               cols: List[Tuple[str, str]],
                                               def_id_col: str):
    for cname, ctype in cols:
        op.execute(sa.text(f'ALTER TABLE {target_table} ADD COLUMN {cname} {ctype}'))
    # ensure that types work properly
    # e.g. heights cannot be text!!
    set_query = ','.join(
        f'{cname} = (SELECT {cname} FROM {Temp.__tablename__} WHERE id = {target_table}.{def_id_col})' for cname, _ in
        cols)
    op.execute(sa.text(f"""
        UPDATE {target_table}
        SET {set_query}
        WHERE EXISTS (SELECT 1 FROM {Temp.__tablename__} WHERE id = {target_table}.{def_id_col});
    """))
    op.execute(sa.text(f"UPDATE {target_table} SET cross_section_width = NULL WHERE cross_section_shape IN (5,6,7)"))
    op.execute(sa.text(f"UPDATE {target_table} SET cross_section_height = NULL WHERE cross_section_shape IN (5,6,7)"))



def migrate_cross_section_definition_to_location():
    cols = [('cross_section_table', 'TEXT'),
            ('cross_section_friction_values', 'TEXT'),
            ('cross_section_vegetation_table', 'TEXT'),
            ('cross_section_shape', 'INT'),
            ('cross_section_width', 'REAL'),
            ('cross_section_height', 'REAL')]
    migrate_cross_section_definition_from_temp(target_table='v2_cross_section_location',
                                               cols=cols,
                                               def_id_col='definition_id')

def migrate_cross_section_definition_to_object(table_name: str):
    cols = [('cross_section_table', 'TEXT'),
            ('cross_section_shape', 'INT'),
            ('cross_section_width', 'REAL'),
            ('cross_section_height', 'REAL')]
    migrate_cross_section_definition_from_temp(target_table=table_name,
                                               cols=cols,
                                               def_id_col='cross_section_definition_id')


def set_geom_for_object(table_name: str, col_name: str = 'the_geom'):
    # line from connection_node_start_id to connection_node_end_id
    # SELECT load_extension('mod_spatialite');
    op.execute(sa.text(f"SELECT AddGeometryColumn('{table_name}', '{col_name}', 4326, 'LINESTRING', 'XY', 0);"))
    q = f"""
        UPDATE
            {table_name}
        SET the_geom = (
            SELECT MakeLine(start_node.the_geom, end_node.the_geom) FROM {table_name} AS object 
            JOIN v2_connection_nodes AS start_node ON object.connection_node_start_id = start_node.id
            JOIN v2_connection_nodes AS end_node ON object.connection_node_end_id = end_node.id
        )         
    """
    op.execute(sa.text(q))


def set_geom_for_v2_pumpstation():
    op.execute(sa.text(f"SELECT AddGeometryColumn('v2_pumpstation', 'the_geom', 4326, 'POINT', 'XY', 0);"))
    q = f"""
        UPDATE
            v2_pumpstation
        SET the_geom = (
            SELECT node.the_geom FROM v2_pumpstation AS object 
            JOIN v2_connection_nodes AS node ON object.connection_node_start_id = node.id
        )         
    """
    op.execute(sa.text(q))


def create_pump_map():
    # Create table
    query = """
        CREATE TABLE pump_map (
                id INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, 
                pump_id INTEGER,connection_node_id_end INTEGER,tags TEXT,code VARCHAR(100),display_name VARCHAR(255),
                geom LINESTRING NOT NULL
            );
    """
    op.execute(sa.text(query))
    # Create geometry
    op.execute(sa.text(f"SELECT AddGeometryColumn('v2_pumpstation', 'map_geom', 4326, 'LINESTRING', 'XY', 0);"))
    op.execute(sa.text("""
        UPDATE v2_pumpstation
        SET map_geom = (
            SELECT MakeLine(start_geom.the_geom, end_geom.the_geom)
            FROM v2_connection_nodes AS start_geom, v2_connection_nodes AS end_geom
            WHERE v2_pumpstation.connection_node_start_id = start_geom.id
            AND v2_pumpstation.connection_node_end_id = end_geom.id
        )
        WHERE EXISTS (
            SELECT 1
            FROM v2_connection_nodes AS start_geom, v2_connection_nodes AS end_geom
            WHERE v2_pumpstation.connection_node_start_id = start_geom.id
            AND v2_pumpstation.connection_node_end_id = end_geom.id
        );
    """))

    # Copy data from v2_pumpstation
    new_col_names = ["pump_id", "connection_node_id_end", "code", "display_name", "geom"]
    old_col_names = ["id", "connection_node_end_id", "code", "display_name", "map_geom"]
    op.execute(sa.text(f"""
        INSERT INTO pump_map ({','.join(new_col_names)}) 
        SELECT {','.join(old_col_names)} FROM v2_pumpstation
        WHERE v2_pumpstation.connection_node_end_id IS NOT NULL
        AND v2_pumpstation.connection_node_start_id IS NOT NULL
    """))


def create_connection_node():
    # Create table
    query = """
            CREATE TABLE connection_node (
            id INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, 
            code VARCHAR(100),tags TEXT,display_name TEXT,storage_area FLOAT,initial_water_level FLOAT,visualisation INTEGER,manhole_surface_level FLOAT,bottom_level FLOAT,exchange_level FLOAT,exchange_type INTEGER,exchange_thickness FLOAT,hydraulic_conductivity_in FLOAT,hydraulic_conductivity_out FLOAT,
            geom POINT NOT NULL
        );    
    """
    op.execute(sa.text(query))
    # copy from v2_connection_nodes
    old_col_names = ["id", "initial_waterlevel", "storage_area", "the_geom", "code"]
    rename_map = {"initial_waterlevel": "initial_water_level", "the_geom": "geom"}
    new_col_names = [rename_map.get(old_name, old_name) for old_name in old_col_names]
    op.execute(sa.text(f"""
        INSERT INTO connection_node ({','.join(new_col_names)}) 
        SELECT {','.join(old_col_names)} FROM v2_connection_nodes
    """))
    # conditional copy from v2_manhole
    old_col_names = ["display_name", "code", "manhole_indicator",
                     "surface_level", "bottom_level", "drain_level",
                     "calculation_type", "exchange_thickness",
                     "hydraulic_conductivity_in", "hydraulic_conductivity_out"]
    rename_map = {"surface_level": "manhole_surface_level",
                  "bottom_level": "bottom_level",
                  "drain_level": "exchange_level",
                  "calculation_type": "exchange_type",
                  "manhole_indicator": "visualisation"}
    set_items = ',\n'.join(f"""{rename_map.get(col_name, col_name)} = (
        SELECT v2_manhole.{col_name} FROM v2_manhole
        WHERE v2_manhole.connection_node_id = connection_node.id)""" for col_name in old_col_names)
    op.execute("CREATE INDEX IF NOT EXISTS idx_v2_manhole_connection_node_id ON v2_manhole(connection_node_id)")
    op.execute(sa.text(f"""
        UPDATE connection_node
        SET {set_items}
        FROM v2_manhole
        WHERE v2_manhole.connection_node_id = connection_node.id;
    """))


# define Material class needed to populate table in create_material
class Material(Base):
    __tablename__ = "material"
    id = Column(Integer, primary_key=True)
    description = Column(Text)
    friction_type = Column(IntegerEnum(constants.FrictionType))
    friction_coefficient = Column(Float)


def create_material():
    op.execute(sa.text("""
    CREATE TABLE material (
    id INTEGER PRIMARY KEY NOT NULL,
    description TEXT,
    friction_type INTEGER,
    friction_coefficient REAL);
    """))
    connection = op.get_bind()
    nof_settings = connection.execute(sa.text("SELECT COUNT(*) FROM model_settings")).scalar()
    session = Session(bind=op.get_bind())
    if nof_settings > 0:
        with open(data_dir.joinpath('0228_materials.csv')) as file:
            reader = csv.DictReader(file)
            session.bulk_save_objects([Material(**row) for row in reader])
            session.commit()


def modify_obstacle():
    op.execute(sa.text(f'ALTER TABLE obstacle ADD COLUMN affects_2d BOOLEAN DEFAULT TRUE;'))
    op.execute(sa.text(f'ALTER TABLE obstacle ADD COLUMN affects_1d2d_open_water BOOLEAN DEFAULT TRUE;'))
    op.execute(sa.text(f'ALTER TABLE obstacle ADD COLUMN affects_1d2d_closed BOOLEAN DEFAULT FALSE;'))


def modify_control_target_type():
    for table_name in ['table_control', 'memory_control']:
        op.execute(sa.text(f"""
            UPDATE {table_name}
            SET target_type = REPLACE(target_type, 'v2_pumpstation', 'pump')
            WHERE target_type = 'v2_pumpstation';
        """))
        op.execute(sa.text(f"""
            UPDATE {table_name}
            SET target_type = REPLACE(target_type, 'v2_', '')
            WHERE target_type LIKE 'v2_%';
        """))



def modify_model_settings():
    op.execute(sa.text(f'ALTER TABLE model_settings ADD COLUMN node_open_water_detection INTEGER DEFAULT 1;'))


def check_for_null_geoms():
    tables = ["v2_connection_nodes", "v2_cross_section_location", "v2_culvert", "v2_channel", "v2_windshielding"]
    conn = op.get_bind()
    for table in tables:
        nof_null = conn.execute(sa.text(f"SELECT COUNT(*) FROM {table} WHERE the_geom IS NULL;")).fetchone()[0]
        if nof_null > 0:
            raise Schema228UpgradeException("Cannot migrate because of empty geometries in table {table}")


def fix_material_id():
    # Replace migrated material_id's with correct values
    replace_map = {9 : 2, 10 : 7}
    material_id_tables = ['pipe', 'culvert', 'weir', 'orifice']
    for table in material_id_tables:
        op.execute(sa.text(f"UPDATE {table} SET material_id = CASE material_id "
                           f"{' '.join([f'WHEN {old} THEN {new}' for old, new in replace_map.items()])} "
                           "ELSE material_id END"))

def upgrade():
    # Empty or non-existing connection node id (start or end) in Orifice, Pipe, Pumpstation or Weir will break
    # migration, so an error is raised in these cases
    check_for_null_geoms()
    # Prevent custom tables in schematisation from breaking migration when they conflict with new table names
    drop_conflicting(op, [new_name for _, new_name in RENAME_TABLES] + ['material', 'pump_map'])
    # Extent cross section definition table (actually stored in temp)
    extend_cross_section_definition_table()
    # Migrate data from cross_section_definition to cross_section_location
    migrate_cross_section_definition_to_location()
    # Prepare object tables for renaming by copying cross section data and setting the_geom
    for table_name in ['v2_culvert', 'v2_weir', 'v2_pipe', 'v2_orifice']:
        migrate_cross_section_definition_to_object(table_name)
        # Set geometry for tables without one
        if table_name != 'v2_culvert':
            set_geom_for_object(table_name)
    set_geom_for_v2_pumpstation()
    for old_table_name, new_table_name in RENAME_TABLES:
        modify_table(old_table_name, new_table_name)
    add_columns_to_tables(ADD_COLUMNS)
    # Create new tables
    create_pump_map()
    create_material()
    create_connection_node()
    # Modify exsiting tables
    modify_model_settings()
    modify_obstacle()
    modify_control_target_type()
    fix_material_id()
    fix_geometry_columns()
    remove_tables([old for old, _ in RENAME_TABLES]+DELETE_TABLES+[Temp.__tablename__, 'v2_manhole'])


def downgrade():
    # Not implemented on purpose
    raise NotImplementedError("Downgrade back from 0.3xx is not supported")



from geoalchemy2.types import Geometry
from sqlalchemy import func, inspect, text
from sqlalchemy.exc import IntegrityError, OperationalError

__all__ = ["copy_models"]


def get_spatialite_version(db):
    with db.session_scope() as session:
        ((lib_version,),) = session.execute(
            text("SELECT spatialite_version()")
        ).fetchall()

    # Identify Spatialite version
    inspector = inspect(db.engine)
    spatial_ref_sys_cols = [x["name"] for x in inspector.get_columns("spatial_ref_sys")]
    if len(spatial_ref_sys_cols) == 0:
        raise ValueError("Not a spatialite file")

    if "srs_wkt" in spatial_ref_sys_cols:
        file_version = 3
    else:
        file_version = 4

    return int(lib_version.split(".")[0]), file_version


def cast_if_geometry(column):
    """Cast Geometry columns to EWKT (so that we can save it right back later)"""
    if isinstance(column.type, Geometry):
        return func.AsEWKT(column).label(column.name)
    else:
        return column


def model_count(session, model) -> int:
    q = text(f"SELECT COUNT(*) FROM {model.__tablename__}")
    try:
        return session.execute(q).fetchall()[0][0]
    except OperationalError as e:
        if "no such table" in str(e):
            return 0
        raise e


def model_query(session, model):
    """Query fields explicitly, so that we end up with an iterator of row tuples"""
    return session.query(*[cast_if_geometry(c) for c in model.__table__.columns])


def model_from_tuple(model, tpl):
    return model(**{c.key: v for (c, v) in zip(model.__table__.columns, tpl)})


def copy_model(from_db, to_db, model):
    with from_db.session_scope() as input_session:
        if model_count(input_session, model) == 0:
            return
        objs = [model_from_tuple(model, x) for x in model_query(input_session, model)]
    with to_db.session_scope() as work_session:
        try:
            work_session.bulk_save_objects(objs)
        except IntegrityError:
            work_session.rollback()
            raise
        else:
            work_session.commit()


def copy_models(from_db, to_db, models):
    for model in models:
        copy_model(from_db, to_db, model)



from sqlalchemy import Boolean, Column, Float, Integer, String, Text
from sqlalchemy.orm import declarative_base

from . import constants
from .custom_types import CSVTable, CSVText, Geometry, IntegerEnum, VarcharEnum

Base = declarative_base()  # automap_base()


class Lateral2D(Base):
    __tablename__ = "lateral_2d"
    id = Column(Integer, primary_key=True)
    code = Column(Text)
    display_name = Column(Text)
    type = Column(IntegerEnum(constants.Later2dType))
    timeseries = Column(CSVText)
    time_units = Column(Text)
    interpolate = Column(Boolean)
    offset = Column(Integer)
    units = Column(Text)
    tags = Column(CSVText)
    geom = Column(Geometry("POINT"), nullable=False)


class BoundaryConditions2D(Base):
    __tablename__ = "boundary_condition_2d"
    id = Column(Integer, primary_key=True)
    code = Column(Text)
    display_name = Column(Text)
    type = Column(IntegerEnum(constants.BoundaryType))
    timeseries = Column(CSVText)
    time_units = Column(Text)
    interpolate = Column(Boolean)
    tags = Column(CSVText)
    geom = Column(Geometry("LINESTRING"), nullable=False)


class ControlMeasureLocation(Base):
    __tablename__ = "measure_location"
    id = Column(Integer, primary_key=True)
    connection_node_id = Column(Integer)
    measure_variable = Column(VarcharEnum(constants.MeasureVariables))
    display_name = Column(Text)
    code = Column(Text)
    geom = Column(Geometry("POINT"), nullable=False)
    tags = Column(CSVText)


class ControlMeasureMap(Base):
    __tablename__ = "measure_map"
    id = Column(Integer, primary_key=True)
    measure_location_id = Column(Integer)
    control_type = Column(VarcharEnum(constants.ControlType), nullable=False)
    control_id = Column(Integer)
    weight = Column(Float)
    display_name = Column(Text)
    code = Column(Text)
    geom = Column(Geometry("LINESTRING"), nullable=False)
    tags = Column(CSVText)


class ControlMemory(Base):
    __tablename__ = "memory_control"
    id = Column(Integer, primary_key=True)
    upper_threshold = Column(Float)
    lower_threshold = Column(Float)
    action_type = Column(VarcharEnum(constants.ControlTableActionTypes))
    action_value_1 = Column(Float)
    action_value_2 = Column(Float)
    target_type = Column(VarcharEnum(constants.StructureControlTypes))
    target_id = Column(Integer)
    is_active = Column(Boolean)
    is_inverse = Column(Boolean)
    display_name = Column(Text)
    code = Column(Text)
    geom = Column(Geometry("POINT"), nullable=False)
    tags = Column(CSVText)


class ControlTable(Base):
    __tablename__ = "table_control"
    id = Column(Integer, primary_key=True)
    action_table = Column(CSVTable)
    action_type = Column(VarcharEnum(constants.ControlTableActionTypes))
    measure_operator = Column(VarcharEnum(constants.MeasureOperators))
    target_type = Column(VarcharEnum(constants.StructureControlTypes))
    target_id = Column(Integer, nullable=False)
    display_name = Column(Text)
    code = Column(Text)
    geom = Column(Geometry("POINT"), nullable=False)
    tags = Column(CSVText)


class Interflow(Base):
    __tablename__ = "interflow"
    id = Column(Integer, primary_key=True)
    interflow_type = Column(IntegerEnum(constants.InterflowType))
    porosity = Column(Float)
    porosity_file = Column(String(255))
    porosity_layer_thickness = Column(Float)
    impervious_layer_elevation = Column(Float)
    hydraulic_conductivity = Column(Float)
    hydraulic_conductivity_file = Column(String(255))


class SimpleInfiltration(Base):
    __tablename__ = "simple_infiltration"
    id = Column(Integer, primary_key=True)
    infiltration_rate = Column(Float)
    infiltration_rate_file = Column(String(255))
    infiltration_surface_option = Column(
        IntegerEnum(constants.InfiltrationSurfaceOption)
    )
    max_infiltration_volume = Column(Float)
    max_infiltration_volume_file = Column(Text)

    # Alias needed for API compatibility
    @property
    def max_infiltration_capacity_file(self):
        return self.max_infiltration_volume_file


class SurfaceParameter(Base):
    __tablename__ = "surface_parameters"
    id = Column(Integer, primary_key=True)
    outflow_delay = Column(Float, nullable=False)
    surface_layer_thickness = Column(Float, nullable=False)
    infiltration = Column(Boolean, nullable=False)
    max_infiltration_capacity = Column(Float, nullable=False)
    min_infiltration_capacity = Column(Float, nullable=False)
    infiltration_decay_constant = Column(Float, nullable=False)
    infiltration_recovery_constant = Column(Float, nullable=False)
    tags = Column(CSVText)
    description = Column(Text)


class Surface(Base):
    __tablename__ = "surface"
    id = Column(Integer, primary_key=True)
    code = Column(String(100))
    display_name = Column(String(255))
    area = Column(Float)
    surface_parameters_id = Column(Integer)
    geom = Column(
        Geometry("POLYGON"),
        nullable=True,
    )
    tags = Column(CSVText)


class DryWeatherFlow(Base):
    __tablename__ = "dry_weather_flow"
    id = Column(Integer, primary_key=True)
    multiplier = Column(Float)
    dry_weather_flow_distribution_id = Column(Integer)
    daily_total = Column(Float)
    interpolate = Column(Boolean)
    display_name = Column(String(255))
    code = Column(String(100))
    geom = Column(
        Geometry("POLYGON"),
        nullable=False,
    )
    tags = Column(CSVText)


class DryWeatherFlowMap(Base):
    __tablename__ = "dry_weather_flow_map"
    id = Column(Integer, primary_key=True)
    connection_node_id = Column(Integer)
    dry_weather_flow_id = Column(Integer)
    display_name = Column(String(255))
    code = Column(String(100))
    geom = Column(
        Geometry("LINESTRING"),
        nullable=False,
    )
    percentage = Column(Float)
    tags = Column(CSVText)


class DryWeatherFlowDistribution(Base):
    __tablename__ = "dry_weather_flow_distribution"
    id = Column(Integer, primary_key=True)
    description = Column(Text)
    tags = Column(CSVText)
    distribution = Column(CSVText)


class GroundWater(Base):
    __tablename__ = "groundwater"
    id = Column(Integer, primary_key=True)

    groundwater_impervious_layer_level = Column(Float)
    groundwater_impervious_layer_level_file = Column(String(255))
    groundwater_impervious_layer_level_aggregation = Column(
        IntegerEnum(constants.InitializationType)
    )
    phreatic_storage_capacity = Column(Float)
    phreatic_storage_capacity_file = Column(String(255))
    phreatic_storage_capacity_aggregation = Column(
        IntegerEnum(constants.InitializationType)
    )
    equilibrium_infiltration_rate = Column(Float)
    equilibrium_infiltration_rate_file = Column(String(255))
    equilibrium_infiltration_rate_aggregation = Column(
        IntegerEnum(constants.InitializationType)
    )
    initial_infiltration_rate = Column(Float)
    initial_infiltration_rate_file = Column(String(255))
    initial_infiltration_rate_aggregation = Column(
        IntegerEnum(constants.InitializationType)
    )
    infiltration_decay_period = Column(Float)
    infiltration_decay_period_file = Column(String(255))
    infiltration_decay_period_aggregation = Column(
        IntegerEnum(constants.InitializationType)
    )
    groundwater_hydraulic_conductivity = Column(Float)
    groundwater_hydraulic_conductivity_file = Column(String(255))
    groundwater_hydraulic_conductivity_aggregation = Column(
        IntegerEnum(constants.InitializationType)
    )
    leakage = Column(Float)
    leakage_file = Column(String(255))

    # Alias needed for API compatibility
    @property
    def groundwater_hydro_connectivity_file(self):
        return self.groundwater_hydraulic_conductivity_file


class GridRefinementLine(Base):
    __tablename__ = "grid_refinement_line"
    id = Column(Integer, primary_key=True)
    display_name = Column(String(255))
    grid_level = Column(Integer)
    geom = Column(Geometry("LINESTRING"), nullable=False)
    code = Column(String(100))
    tags = Column(CSVText)


class GridRefinementArea(Base):
    __tablename__ = "grid_refinement_area"
    id = Column(Integer, primary_key=True)
    display_name = Column(String(255))
    grid_level = Column(Integer)
    code = Column(String(100))
    geom = Column(Geometry("POLYGON"), nullable=False)
    tags = Column(CSVText)


class ConnectionNode(Base):
    __tablename__ = "connection_node"
    id = Column(Integer, primary_key=True)
    geom = Column(Geometry("POINT"), nullable=False)
    code = Column(String(100))
    tags = Column(CSVText)
    display_name = Column(Text)
    storage_area = Column(Float)
    initial_water_level = Column(Float)
    visualisation = Column(Integer)
    manhole_surface_level = Column(Float)
    bottom_level = Column(Float)
    exchange_level = Column(Float)
    exchange_type = Column(IntegerEnum(constants.CalculationTypeNode))
    exchange_thickness = Column(Float)
    hydraulic_conductivity_in = Column(Float)
    hydraulic_conductivity_out = Column(Float)


class Lateral1d(Base):
    __tablename__ = "lateral_1d"
    id = Column(Integer, primary_key=True)
    code = Column(Text)
    display_name = Column(Text)
    timeseries = Column(CSVText)
    time_units = Column(Text)
    interpolate = Column(Boolean)
    offset = Column(Integer)
    units = Column(Text)
    tags = Column(CSVText)
    geom = Column(Geometry("POINT"), nullable=False)

    connection_node_id = Column(Integer)


class NumericalSettings(Base):
    __tablename__ = "numerical_settings"
    id = Column(Integer, primary_key=True)
    cfl_strictness_factor_1d = Column(Float)
    cfl_strictness_factor_2d = Column(Float)
    convergence_cg = Column(Float)
    convergence_eps = Column(Float)
    flow_direction_threshold = Column(Float)
    friction_shallow_water_depth_correction = Column(
        IntegerEnum(constants.FrictionShallowWaterDepthCorrection)
    )
    general_numerical_threshold = Column(Float)
    time_integration_method = Column(IntegerEnum(constants.IntegrationMethod))
    limiter_waterlevel_gradient_1d = Column(IntegerEnum(constants.OffOrStandard))
    limiter_waterlevel_gradient_2d = Column(IntegerEnum(constants.OffOrStandard))
    limiter_slope_crossectional_area_2d = Column(
        IntegerEnum(constants.LimiterSlopeXArea)
    )
    limiter_slope_friction_2d = Column(IntegerEnum(constants.OffOrStandard))
    max_non_linear_newton_iterations = Column(Integer)
    max_degree_gauss_seidel = Column(Integer)
    min_friction_velocity = Column(Float)
    min_surface_area = Column(Float)
    use_preconditioner_cg = Column(IntegerEnum(constants.OffOrStandard))
    preissmann_slot = Column(Float)
    pump_implicit_ratio = Column(Float)
    limiter_slope_thin_water_layer = Column(Float)
    use_of_cg = Column(Integer)
    use_nested_newton = Column(IntegerEnum(constants.OffOrStandard))
    flooding_threshold = Column(Float)


class VegetationDrag(Base):
    __tablename__ = "vegetation_drag_2d"
    id = Column(Integer, primary_key=True)

    vegetation_height = Column(Float)
    vegetation_height_file = Column(String(255))

    vegetation_stem_count = Column(Float)
    vegetation_stem_count_file = Column(String(255))

    vegetation_stem_diameter = Column(Float)
    vegetation_stem_diameter_file = Column(String(255))

    vegetation_drag_coefficient = Column(Float)
    vegetation_drag_coefficient_file = Column(String(255))


class ModelSettings(Base):
    __tablename__ = "model_settings"
    id = Column(Integer, primary_key=True)
    use_2d_flow = Column(Boolean)
    use_1d_flow = Column(Boolean)
    manhole_aboveground_storage_area = Column(Float)
    minimum_cell_size = Column(Float)
    calculation_point_distance_1d = Column(Float)
    nr_grid_levels = Column(Integer)
    minimum_table_step_size = Column(Float)
    maximum_table_step_size = Column(Float)
    dem_file = Column(String(255))
    friction_type = Column(IntegerEnum(constants.FrictionType))
    friction_coefficient = Column(Float)
    friction_coefficient_file = Column(String(255))
    embedded_cutoff_threshold = Column(Float)
    max_angle_1d_advection = Column(Float)
    friction_averaging = Column(Boolean)
    table_step_size_1d = Column(Float)
    use_2d_rain = Column(Boolean)
    use_interflow = Column(Boolean)
    use_interception = Column(Boolean)
    use_simple_infiltration = Column(Boolean)
    use_groundwater_flow = Column(Boolean)
    use_groundwater_storage = Column(Boolean)
    use_vegetation_drag_2d = Column(Boolean)
    node_open_water_detection = Column(IntegerEnum(constants.NodeOpenWaterDetection))

    # Alias needed for API compatibility
    @property
    def frict_coef_file(self):
        return self.friction_coefficient_file


class InitialConditions(Base):
    __tablename__ = "initial_conditions"
    id = Column(Integer, primary_key=True)
    initial_groundwater_level = Column(Float)
    initial_groundwater_level_file = Column(String(255))
    initial_groundwater_level_aggregation = Column(
        IntegerEnum(constants.InitializationType)
    )
    initial_water_level = Column(Float)
    initial_water_level_aggregation = Column(IntegerEnum(constants.InitializationType))
    initial_water_level_file = Column(String(255))

    # Alias needed for API compatibility
    @property
    def initial_waterlevel_file(self):
        return self.initial_water_level_file


class Interception(Base):
    __tablename__ = "interception"
    id = Column(Integer, primary_key=True)
    interception = Column(Float)
    interception_file = Column(String(255))


# class PhysicalSettings
class AggregationSettings(Base):
    __tablename__ = "aggregation_settings"
    id = Column(Integer, primary_key=True)
    flow_variable = Column(VarcharEnum(constants.FlowVariable))
    aggregation_method = Column(VarcharEnum(constants.AggregationMethod))
    interval = Column(Integer)


class PhysicalSettings(Base):
    __tablename__ = "physical_settings"
    id = Column(Integer, primary_key=True)
    use_advection_1d = Column(IntegerEnum(constants.AdvectionTypes1D))
    use_advection_2d = Column(Boolean)


class SimulationTemplateSettings(Base):
    __tablename__ = "simulation_template_settings"
    id = Column(Integer, primary_key=True)
    name = Column(String(128))
    use_0d_inflow = Column(IntegerEnum(constants.InflowType))
    use_structure_control = Column(Boolean)


class TimeStepSettings(Base):
    __tablename__ = "time_step_settings"
    id = Column(Integer, primary_key=True)
    time_step = Column(Float)
    min_time_step = Column(Float)
    max_time_step = Column(Float)
    output_time_step = Column(Float)
    use_time_step_stretch = Column(Boolean)


class BoundaryCondition1D(Base):
    __tablename__ = "boundary_condition_1d"

    id = Column(Integer, primary_key=True)
    code = Column(Text)
    display_name = Column(Text)
    type = Column(IntegerEnum(constants.BoundaryType))
    timeseries = Column(CSVText)
    time_units = Column(Text)
    interpolate = Column(Boolean)
    tags = Column(CSVText)
    geom = Column(Geometry("POINT"), nullable=False)

    connection_node_id = Column(Integer)


class SurfaceMap(Base):
    __tablename__ = "surface_map"
    id = Column(Integer, primary_key=True)
    surface_id = Column(Integer, nullable=False)
    connection_node_id = Column(Integer)
    percentage = Column(Float)
    geom = Column(Geometry("LINESTRING"), nullable=False)
    tags = Column(CSVText)
    code = Column(String(100))
    display_name = Column(String(255))


class Channel(Base):
    __tablename__ = "channel"
    id = Column(Integer, primary_key=True)
    display_name = Column(String(255))
    code = Column(String(100))
    tags = Column(CSVText)
    exchange_type = Column(IntegerEnum(constants.CalculationType))
    calculation_point_distance = Column(Float)
    geom = Column(Geometry("LINESTRING"), nullable=False)
    connection_node_id_start = Column(Integer)
    connection_node_id_end = Column(Integer)
    exchange_thickness = Column(Float)
    hydraulic_conductivity_in = Column(Float)
    hydraulic_conductivity_out = Column(Float)


class Windshielding(Base):
    __tablename__ = "windshielding_1d"
    id = Column(Integer, primary_key=True)
    north = Column(Float)
    northeast = Column(Float)
    east = Column(Float)
    southeast = Column(Float)
    south = Column(Float)
    southwest = Column(Float)
    west = Column(Float)
    northwest = Column(Float)
    geom = Column(Geometry("POINT"), nullable=False)
    channel_id = Column(Integer)
    tags = Column(CSVText)


class CrossSectionLocation(Base):
    __tablename__ = "cross_section_location"
    id = Column(Integer, primary_key=True)
    code = Column(String(100))
    tags = Column(CSVText)
    reference_level = Column(Float)
    friction_type = Column(IntegerEnum(constants.FrictionType))
    friction_value = Column(Float)
    bank_level = Column(Float)
    cross_section_shape = Column(IntegerEnum(constants.CrossSectionShape))
    cross_section_width = Column(Float)
    cross_section_height = Column(Float)
    cross_section_friction_values = Column(CSVText)
    cross_section_vegetation_table = Column(CSVTable)
    cross_section_table = Column(CSVTable)
    vegetation_stem_density = Column(Float)
    vegetation_stem_diameter = Column(Float)
    vegetation_height = Column(Float)
    vegetation_drag_coefficient = Column(Float)
    geom = Column(Geometry("POINT"), nullable=False)
    channel_id = Column(Integer)


class Pipe(Base):
    __tablename__ = "pipe"
    id = Column(Integer, primary_key=True)
    display_name = Column(String(255))
    code = Column(String(100))
    tags = Column(CSVText)
    geom = Column(Geometry("LINESTRING"), nullable=False)
    sewerage_type = Column(IntegerEnum(constants.SewerageType))
    exchange_type = Column(IntegerEnum(constants.PipeCalculationType))
    invert_level_start = Column(Float)
    invert_level_end = Column(Float)
    friction_value = Column(Float)
    friction_type = Column(IntegerEnum(constants.FrictionType))
    calculation_point_distance = Column(Float)
    material_id = Column(Integer)
    connection_node_id_start = Column(Integer)
    connection_node_id_end = Column(Integer)
    cross_section_shape = Column(IntegerEnum(constants.CrossSectionShape))
    cross_section_width = Column(Float)
    cross_section_height = Column(Float)
    cross_section_table = Column(CSVTable)
    exchange_thickness = Column(Float)
    hydraulic_conductivity_in = Column(Float)
    hydraulic_conductivity_out = Column(Float)


class Culvert(Base):
    __tablename__ = "culvert"
    id = Column(Integer, primary_key=True)
    display_name = Column(String(255))
    code = Column(String(100))
    tags = Column(CSVText)
    exchange_type = Column(IntegerEnum(constants.CalculationTypeCulvert))
    friction_value = Column(Float)
    friction_type = Column(IntegerEnum(constants.FrictionType))
    calculation_point_distance = Column(Float)
    discharge_coefficient_positive = Column(Float)
    discharge_coefficient_negative = Column(Float)
    invert_level_start = Column(Float)
    invert_level_end = Column(Float)
    geom = Column(Geometry("LINESTRING"), nullable=False)
    material_id = Column(Integer)
    connection_node_id_start = Column(Integer)
    connection_node_id_end = Column(Integer)
    cross_section_shape = Column(IntegerEnum(constants.CrossSectionShape))
    cross_section_width = Column(Float)
    cross_section_height = Column(Float)
    cross_section_table = Column(CSVTable)


class DemAverageArea(Base):
    __tablename__ = "dem_average_area"
    id = Column(Integer, primary_key=True)
    geom = Column(Geometry("POLYGON"), nullable=False)
    display_name = Column(Text)
    code = Column(Text)
    tags = Column(CSVText)


class Weir(Base):
    __tablename__ = "weir"
    id = Column(Integer, primary_key=True)
    code = Column(String(100))
    display_name = Column(String(255))
    geom = Column(Geometry("LINESTRING"), nullable=False)
    tags = Column(CSVText)
    crest_level = Column(Float)
    crest_type = Column(IntegerEnum(constants.CrestType))
    friction_value = Column(Float)
    friction_type = Column(IntegerEnum(constants.FrictionType))
    discharge_coefficient_positive = Column(Float)
    discharge_coefficient_negative = Column(Float)
    material_id = Column(Integer)
    sewerage = Column(Boolean)
    external = Column(Boolean)
    connection_node_id_start = Column(Integer)
    connection_node_id_end = Column(Integer)
    cross_section_shape = Column(IntegerEnum(constants.CrossSectionShape))
    cross_section_width = Column(Float)
    cross_section_height = Column(Float)
    cross_section_table = Column(CSVTable)


class Orifice(Base):
    __tablename__ = "orifice"
    id = Column(Integer, primary_key=True)
    code = Column(String(100))
    display_name = Column(String(255))
    tags = Column(CSVText)
    geom = Column(Geometry("LINESTRING"), nullable=False)
    crest_type = Column(IntegerEnum(constants.CrestType))
    crest_level = Column(Float)
    material_id = Column(Integer)
    friction_value = Column(Float)
    friction_type = Column(IntegerEnum(constants.FrictionType))
    discharge_coefficient_positive = Column(Float)
    discharge_coefficient_negative = Column(Float)
    sewerage = Column(Boolean)
    connection_node_id_start = Column(Integer)
    connection_node_id_end = Column(Integer)
    cross_section_shape = Column(IntegerEnum(constants.CrossSectionShape))
    cross_section_width = Column(Float)
    cross_section_height = Column(Float)
    cross_section_table = Column(CSVTable)


class Pump(Base):
    __tablename__ = "pump"
    id = Column(Integer, primary_key=True)
    code = Column(String(100))
    display_name = Column(String(255))
    start_level = Column(Float)
    lower_stop_level = Column(Float)
    upper_stop_level = Column(Float)
    capacity = Column(Float)
    type_ = Column(
        IntegerEnum(constants.PumpType), name="type", key="type_"
    )  # type: ignore[call-overload]
    sewerage = Column(Boolean)
    connection_node_id = Column(Integer)
    geom = Column(Geometry("POINT"), nullable=False)
    tags = Column(CSVText)


class PumpMap(Base):
    __tablename__ = "pump_map"
    id = Column(Integer, primary_key=True)
    pump_id = Column(Integer)
    connection_node_id_end = Column(Integer)
    geom = Column(Geometry("LINESTRING"), nullable=False)
    tags = Column(CSVText)
    code = Column(String(100))
    display_name = Column(String(255))


class Obstacle(Base):
    __tablename__ = "obstacle"
    id = Column(Integer, primary_key=True)
    code = Column(String(100))
    crest_level = Column(Float)
    geom = Column(Geometry("LINESTRING"), nullable=False)
    tags = Column(CSVText)
    display_name = Column(String(255))
    affects_2d = Column(Boolean)
    affects_1d2d_open_water = Column(Boolean)
    affects_1d2d_closed = Column(Boolean)


class PotentialBreach(Base):
    __tablename__ = "potential_breach"
    id = Column(Integer, primary_key=True)
    code = Column(String(100))
    display_name = Column(String(255))
    tags = Column(CSVText)
    initial_exchange_level = Column(Float)
    final_exchange_level = Column(Float)
    levee_material = Column(IntegerEnum(constants.Material))
    geom = Column(Geometry("LINESTRING"), nullable=False)
    channel_id = Column(Integer)


class ExchangeLine(Base):
    __tablename__ = "exchange_line"
    id = Column(Integer, primary_key=True)
    geom = Column(Geometry("LINESTRING"), nullable=False)
    channel_id = Column(Integer)
    exchange_level = Column(Float)
    display_name = Column(Text)
    code = Column(Text)
    tags = Column(CSVText)


class Tags(Base):
    __tablename__ = "tag"
    id = Column(Integer, primary_key=True)
    description = Column(Text)


class Material(Base):
    __tablename__ = "material"
    id = Column(Integer, primary_key=True)
    description = Column(Text)
    friction_type = Column(IntegerEnum(constants.FrictionType))
    friction_coefficient = Column(Float)


DECLARED_MODELS = [
    AggregationSettings,
    BoundaryCondition1D,
    BoundaryConditions2D,
    Channel,
    ConnectionNode,
    ControlMeasureLocation,
    ControlMeasureMap,
    ControlMemory,
    ControlTable,
    CrossSectionLocation,
    Culvert,
    DemAverageArea,
    DryWeatherFlow,
    DryWeatherFlowMap,
    DryWeatherFlowDistribution,
    ExchangeLine,
    GridRefinementLine,
    GridRefinementArea,
    GroundWater,
    InitialConditions,
    Interception,
    Interflow,
    Lateral1d,
    Lateral2D,
    Material,
    ModelSettings,
    NumericalSettings,
    Obstacle,
    Orifice,
    PhysicalSettings,
    Pipe,
    PotentialBreach,
    Pump,
    PumpMap,
    SimpleInfiltration,
    SimulationTemplateSettings,
    Surface,
    SurfaceMap,
    SurfaceParameter,
    Tags,
    TimeStepSettings,
    VegetationDrag,
    Weir,
    Windshielding,
]



"""vegetation drag

Revision ID: 0216
Revises: 0215
Create Date: 2023-03-13 15:05:42.393961

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '0216'
down_revision = '0215'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'v2_vegetation_drag',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('display_name', sa.String(length=255), nullable=True),
        sa.Column('height', sa.Float(), nullable=True),
        sa.Column('height_file', sa.String(length=255), nullable=True),
        sa.Column('stem_count', sa.Float(), nullable=True),
        sa.Column('stem_count_file', sa.String(length=255), nullable=True),
        sa.Column('stem_diameter', sa.Float(), nullable=True),
        sa.Column('stem_diameter_file', sa.String(length=255), nullable=True),
        sa.Column('drag_coefficient', sa.Float(), nullable=True),
        sa.Column('drag_coefficient_file', sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )

    op.add_column('v2_global_settings', sa.Column('vegetation_drag_settings_id', sa.Integer(), nullable=True))

    ## FIX migration 214:
    for table_name in ["v2_connected_pnt", "v2_calculation_point", "v2_levee"]:
        try:
            op.execute(sa.text(f"SELECT DropTable(NULL, '{table_name}', TRUE)"))
        except sa.exc.OperationalError:
            op.execute(sa.text(f"SELECT DropGeoTable('{table_name}')"))
            op.execute(sa.text(f"DROP TABLE IF EXISTS '{table_name}'"))


def downgrade():
    op.drop_constraint(None, 'v2_global_settings', type_='foreignkey')
    op.drop_column('v2_global_settings', 'vegetation_drag_settings_id')
    op.drop_table('v2_vegetation_drag')



# comparing directly with SQLAlchemy Column type doesn't work so we use InstrumentedAttribute
from sqlalchemy.orm.attributes import InstrumentedAttribute

# this import is a test in itself; if the beta variables aren't defined, the modelchecker will fail
from threedi_schema.beta_features import BETA_COLUMNS, BETA_VALUES


def test_beta_columns_structure():
    assert isinstance(BETA_COLUMNS, list)
    for column in BETA_COLUMNS:
        assert isinstance(column, InstrumentedAttribute)


def test_beta_values_structure():
    assert isinstance(BETA_VALUES, list)
    for entry in BETA_VALUES:
        assert set(entry) == {"columns", "values"}  # check the keys
        assert isinstance(entry["columns"], list)
        assert isinstance(entry["values"], list)
        # empty columns or values field is not allowed
        assert len(entry["columns"]) > 0
        assert len(entry["values"]) > 0
        for column in entry["columns"]:
            assert isinstance(column, InstrumentedAttribute)



from .domain import constants, models  # noqa

"""
Put features in beta development in these lists to prevent users from using them, in the SQLAlchemy format used in the schema.
"""

BETA_COLUMNS = []
"""This list contains beta columns, e.g. models.SomeTable.some_column"""

BETA_VALUES = []
"""
This list contains dicts with lists of beta values for columns, where the dict has the format:
{
    "columns": [
        models.SomeTable.some_column,
        models.SomeTable.some_other_column,
    ],
    "values": [
        constants.SomeConstantsClass.SOMEENUMVALUE,
        "or just some random string",
    ]
}
The modelchecker will go through each column and give an error when a beta value is used for its associated column.
"""
